# -*- coding: utf-8 -*-
#  Copyright 2021 United Kingdom Research and Innovation
#  Copyright 2021 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import unittest
from cil.framework import ImageGeometry, AcquisitionGeometry
import numpy as np
from utils import has_tomophantom, initialise_tests

initialise_tests()

if has_tomophantom:
    from cil.plugins import TomoPhantom


class TestTomoPhantom2D(unittest.TestCase):
    def setUp(self):
        
        N=128
        angles = np.linspace(0, 360, 50, True, dtype=np.float32)
        offset = 0.4
        ag = AcquisitionGeometry.create_Cone2D((offset,-100), (offset,100))
        ag.set_panel(N)
        
        ag.set_angles(angles, angle_unit=AcquisitionGeometry.DEGREE)
        ig = ag.get_ImageGeometry()
        self.ag = ag
        self.ig = ig
        self.N = N

    def tearDown(self):
        pass
    @unittest.skipUnless(has_tomophantom, 'Please install TomoPhantom')
    def test_2D(self):
        model = 1

        phantom = TomoPhantom.get_ImageData(model, self.ig)

        assert phantom.geometry.channels == 1
        assert phantom.shape == (self.N,self.N)

    @unittest.skipUnless(has_tomophantom, 'Please install TomoPhantom')
    def test_MC2D_wrong_model(self):
        model = 1
        ag = self.ag.copy()
        ag.set_channels(3)
        ig = ag.get_ImageGeometry()
        try:
            phantom = TomoPhantom.get_ImageData(model, ig)
            assert False
        except ValueError as ve:
            assert True
    @unittest.skipUnless(has_tomophantom, 'Please install TomoPhantom')
    def test_MC2D_wrong_amount_channels(self):
        model = 1
        ag = self.ag.copy()
        ag.set_channels(3)
        ig = ag.get_ImageGeometry()
        try:
            phantom = TomoPhantom.get_ImageData(model, ig)
            assert False
        except ValueError as ve:
            assert True
    @unittest.skipUnless(has_tomophantom, 'Please install TomoPhantom')
    def test_MC2D(self):
        model = 100
        self.ig.channels =3
        phantom = TomoPhantom.get_ImageData(model, self.ig)
        assert phantom.geometry.channels == self.ig.channels
        assert phantom.shape == self.ig.shape

    @unittest.skipUnless(has_tomophantom, 'Please install TomoPhantom')
    def test_is_model_temporal(self):
        model = 100
        assert TomoPhantom.is_model_temporal(model, num_dims=2)

    @unittest.skipUnless(has_tomophantom, 'Please install TomoPhantom')
    def test_get_model_num_channels(self):
        model = 100
        N = TomoPhantom.is_model_temporal(model, num_dims=2)
        assert N >= 1

class TestTomoPhantom3D(unittest.TestCase):
    def setUp(self):
        
        N=128
        angles = np.linspace(0, 360, 50, True, dtype=np.float32)
        offset = 0.4
        ag = AcquisitionGeometry.create_Cone3D((offset,-100,0), (offset,100,0))
        ag.set_panel((N,N/2))
        
        ag.set_angles(angles, angle_unit=AcquisitionGeometry.DEGREE)
        ig = ag.get_ImageGeometry()
        self.ag = ag
        self.ig = ig
        self.N = N

    def tearDown(self):
        pass
    @unittest.skipUnless(has_tomophantom, 'Please install TomoPhantom')
    def test_3D(self):
        model = 1

        phantom = TomoPhantom.get_ImageData(model, self.ig)
        assert phantom.geometry.channels == 1
        assert phantom.shape == (self.N/2,self.N, self.N)

    @unittest.skipUnless(has_tomophantom, 'Please install TomoPhantom')
    def test_MC3D_wrong_model(self):
        model = 1
        ag = self.ag.copy()
        ag.set_channels(3)
        ig = ag.get_ImageGeometry()
        try:
            phantom = TomoPhantom.get_ImageData(model, ig)
            assert False
        except ValueError as ve:
            assert True
    @unittest.skipUnless(has_tomophantom, 'Please install TomoPhantom')
    def test_MC3D_wrong_amount_channels(self):
        model = 1
        ag = self.ag.copy()
        ag.set_channels(3)
        ig = ag.get_ImageGeometry()
        try:
            phantom = TomoPhantom.get_ImageData(model, ig)
            assert False
        except ValueError as ve:
            assert True
    @unittest.skipUnless(has_tomophantom, 'Please install TomoPhantom')
    def test_MC3D(self):
        model = 100
        ag = self.ag.copy()
        ag.set_channels(5)
        ig = ag.get_ImageGeometry()
        phantom = TomoPhantom.get_ImageData(model, ig)
        
        assert phantom.geometry.channels == ig.channels
        assert phantom.shape == ig.shape
    @unittest.skipUnless(has_tomophantom, 'Please install TomoPhantom')
    def test_is_model_temporal(self):
        model = 100
        assert TomoPhantom.is_model_temporal(model, num_dims=3)
    @unittest.skipUnless(has_tomophantom, 'Please install TomoPhantom')
    def test_get_model_num_channels(self):
        model = 100
        N = TomoPhantom.is_model_temporal(model, num_dims=3)
        assert N >= 1



# -*- coding: utf-8 -*-
#  Copyright 2020 United Kingdom Research and Innovation
#  Copyright 2020 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.framework import DataOrder
from cil.optimisation.operators import LinearOperator, ChannelwiseOperator
from cil.framework.BlockGeometry import BlockGeometry
from cil.optimisation.operators import BlockOperator
from cil.plugins.astra.operators import AstraProjector3D
from cil.plugins.astra.operators import AstraProjector2D
import logging


class ProjectionOperator(LinearOperator):
    """
    ProjectionOperator configures and calls appropriate ASTRA Projectors for your dataset.

    Parameters
    ----------

    image_geometry : ``ImageGeometry``, default used if None
        A description of the area/volume to reconstruct

    acquisition_geometry : ``AcquisitionGeometry``, ``BlockGeometry``
        A description of the acquisition data. If passed a BlockGeometry it will return a BlockOperator.

    device : string, default='gpu'
        'gpu' will run on a compatible CUDA capable device using the ASTRA 3D CUDA Projectors, 'cpu' will run on CPU using the ASTRA 2D CPU Projectors

    Example
    -------
    >>> from cil.plugins.astra import ProjectionOperator
    >>> PO = ProjectionOperator(image.geometry, data.geometry)
    >>> forward_projection = PO.direct(image)
    >>> backward_projection = PO.adjoint(data)

    Notes
    -----
    For multichannel data the ProjectionOperator will broadcast across all channels.
    """
    def __new__(cls, image_geometry=None, acquisition_geometry=None, \
        device='gpu', **kwargs):
        if isinstance(acquisition_geometry, BlockGeometry):
            logging.info("BlockOperator is returned.")

            K = []
            for ag in acquisition_geometry:
                K.append(
                    ProjectionOperator_ag(image_geometry=image_geometry, acquisition_geometry=ag, \
                        device=device, **kwargs)
                )
            return BlockOperator(*K)
        else:
            logging.info("Standard Operator is returned.")
            return super(ProjectionOperator,
                         cls).__new__(ProjectionOperator_ag)


class ProjectionOperator_ag(ProjectionOperator):
    """
    ProjectionOperator configures and calls appropriate ASTRA Projectors for your dataset.

    Parameters
    ----------

    image_geometry : ImageGeometry, default used if None
        A description of the area/volume to reconstruct

    acquisition_geometry : AcquisitionGeometry
        A description of the acquisition data

    device : string, default='gpu'
        'gpu' will run on a compatible CUDA capable device using the ASTRA 3D CUDA Projectors, 'cpu' will run on CPU using the ASTRA 2D CPU Projectors

    Example
    -------
    >>> from cil.plugins.astra import ProjectionOperator
    >>> PO = ProjectionOperator(image.geometry, data.geometry)
    >>> forward_projection = PO.direct(image)
    >>> backward_projection = PO.adjoint(data)

    Notes
    -----
    For multichannel data the ProjectionOperator will broadcast across all channels.
    """

    def __init__(self,
                 image_geometry=None,
                 acquisition_geometry=None,
                 device='gpu'):

        if acquisition_geometry is None:
            raise TypeError(
                "Please specify an acquisition_geometry to configure this operator"
            )

        if image_geometry is None:
            image_geometry = acquisition_geometry.get_ImageGeometry()

        super(ProjectionOperator_ag,
              self).__init__(domain_geometry=image_geometry,
                             range_geometry=acquisition_geometry)

        DataOrder.check_order_for_engine('astra', image_geometry)
        DataOrder.check_order_for_engine('astra', acquisition_geometry)

        self.volume_geometry = image_geometry
        self.sinogram_geometry = acquisition_geometry

        sinogram_geometry_sc = acquisition_geometry.get_slice(channel=0)
        volume_geometry_sc = image_geometry.get_slice(channel=0)

        if device == 'gpu':
            operator = AstraProjector3D(volume_geometry_sc,
                                        sinogram_geometry_sc)
        elif self.sinogram_geometry.dimension == '2D':
            operator = AstraProjector2D(volume_geometry_sc,
                                        sinogram_geometry_sc,
                                        device=device)
        else:
            raise NotImplementedError("Cannot process 3D data without a GPU")

        if acquisition_geometry.channels > 1:
            operator_full = ChannelwiseOperator(
                operator, self.sinogram_geometry.channels, dimension='prepend')
            self.operator = operator_full
        else:
            self.operator = operator

    def direct(self, IM, out=None):
        '''Applies the direct of the operator i.e. the forward projection.
        
        Parameters
        ----------
        IM : ImageData
            The image/volume to be projected.

        out : DataContainer, optional
           Fills the referenced DataContainer with the processed data and suppresses the return
        
        Returns
        -------
        DataContainer
            The processed data. Suppressed if `out` is passed
        '''

        return self.operator.direct(IM, out=out)

    def adjoint(self, DATA, out=None):
        '''Applies the adjoint of the operator, i.e. the backward projection.

        Parameters
        ----------
        DATA : AcquisitionData
            The projections/sinograms to be projected.

        out : DataContainer, optional
           Fills the referenced DataContainer with the processed data and suppresses the return
        
        Returns
        -------
        DataContainer
            The processed data. Suppressed if `out` is passed
        '''
        return self.operator.adjoint(DATA, out=out)

    def calculate_norm(self):
        return self.operator.norm()

    def domain_geometry(self):
        return self.volume_geometry

    def range_geometry(self):
        return self.sinogram_geometry



# -*- coding: utf-8 -*-
#  Copyright 2023 United Kingdom Research and Innovation
#  Copyright 2023 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.framework import cilacc

import numpy as np
import ctypes

c_float_p = ctypes.POINTER(ctypes.c_float)
c_size_t_p = ctypes.POINTER(ctypes.c_size_t)

class Binner_IPP(object):

   def __init__(self, shape_in, shape_out, start_index, binning):
      """This constructs a configured cilacc Binner object used by cil.processors.Binner. This class should not be used directly.
        
      This performs no checks on the inputs. Expects input for 4 dimensions.

      Parameters
      ----------

      shape_in : list
         Input array shape. List of 4 integers corresponding to each dimension. i.e. [1,1,512,512]

      shape_out : list
         Output array shape. List of 4 integers corresponding to each dimension. i.e. [1,1,246,246]

      start_index : list
         Starting pixel index for cropping. List of 4 integers corresponding to each dimension. i.e. [0,0,1,1]

      binning : list
         Number of pixels to average together in each direction. List of 4 integers corresponding to each dimension. i.e. [1,1,2,2]

      """

      cilacc.Binner_new.argtypes = [c_size_t_p,c_size_t_p,c_size_t_p,c_size_t_p]
      cilacc.Binner_new.restype = ctypes.c_void_p

      cilacc.Binner_bin.argtypes =  [ ctypes.c_void_p, c_float_p,c_float_p]
      cilacc.Binner_bin.restype = ctypes.c_int32

      cilacc.Binner_delete.argtypes =  [ ctypes.c_void_p]
      cilacc.Binner_delete.restype = ctypes.c_void_p

      shape_in_arr = np.array(shape_in, np.uintp)
      shape_out_arr = np.array(shape_out, np.uintp)
      start_index_arr = np.array(start_index, np.uintp)
      binning_arr = np.array(binning, np.uintp)

      shape_in_p = shape_in_arr.ctypes.data_as(c_size_t_p)
      shape_out_p = shape_out_arr.ctypes.data_as(c_size_t_p)
      ind_start_p = start_index_arr.ctypes.data_as(c_size_t_p)
      binning_p = binning_arr.ctypes.data_as(c_size_t_p)

      for i in range(4):
         if shape_out_p[i] * binning_p[i] + ind_start_p[i] > shape_in_p[i]:
            raise ValueError("Input dimension mismatch on dimension {0}".format(i))
   
      self.obj = cilacc.Binner_new(shape_in_p, shape_out_p, ind_start_p, binning_p)


   def bin(self, array_in, array_binned):
      """This bins the input array and writes the result in array_binned.
      
      This  performs no checks on the inputs.

      Parameters
      ----------

      array_in : ndarray
         Must have shape corresponding to shape_in. Data type float32. C ordered, contiguous memory

      array_binned : ndarray
         Must have shape corresponding to shape_out. Data type float32. C ordered, contiguous memory
      """
      data_p = array_in.ctypes.data_as(c_float_p)
      data_out_p = array_binned.ctypes.data_as(c_float_p)

      return cilacc.Binner_bin(self.obj, data_p, data_out_p)

   def __del__(self):
      """This deletes the cilacc Binner object 
      """

      if hasattr(self,'obj'):
         cilacc.Binner_delete(self.obj)
